#
# ur_decoder.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .ur import UR
from .fountain_encoder import Part as FountainEncoderPart
from .fountain_decoder import FountainDecoder
from .bytewords import *
from .utils import drop_first, is_ur_type


class InvalidScheme(Exception):
    pass


class InvalidType(Exception):
    pass


class InvalidPathLength(Exception):
    pass


class InvalidSequenceComponent(Exception):
    pass


class InvalidFragment(Exception):
    pass


class URDecoder:
    def __init__(self):
        self.fountain_decoder = FountainDecoder()
        self.expected_type = None
        self.result = None

    @staticmethod
    def decode(str):
        (type, components) = URDecoder.parse(str)
        if len(components) == 0:
            raise InvalidPathLength()

        body = components[0]
        return URDecoder.decode_by_type(type, body)

    @staticmethod
    def decode_by_type(type, body):
        cbor = Bytewords.decode(Bytewords_Style_minimal, body)
        return UR(type, cbor)

    @staticmethod
    def parse(str):
        # Don't consider case
        lowered = str.lower()

        # Validate URI scheme
        if not lowered.startswith("ur:"):
            raise InvalidScheme()

        path = drop_first(lowered, 3)

        # Split the remainder into path components
        components = path.split("/")

        # Make sure there are at least two path components
        if len(components) < 2:
            raise InvalidPathLength()

        # Validate the type
        type = components[0]
        if not is_ur_type(type):
            raise InvalidType()

        comps = components[1:]  # Don't include the ur type
        return (type, comps)

    @staticmethod
    def parse_sequence_component(str):
        try:
            comps = str.split("-")
            if len(comps) != 2:
                raise InvalidSequenceComponent()
            seq_num = int(comps[0])
            seq_len = int(comps[1])
            if seq_num < 1 or seq_len < 1:
                raise InvalidSequenceComponent()
            return (seq_num, seq_len)
        except:
            raise InvalidSequenceComponent()

    def validate_part(self, type):
        if self.expected_type == None:
            if not is_ur_type(type):
                return False
            self.expected_type = type
            return True
        else:
            return type == self.expected_type

    def receive_part(self, str):
        try:
            # Don't process the part if we're already done
            if self.result != None:
                return False

            # Don't continue if this part doesn't validate
            (type, components) = URDecoder.parse(str)
            if not self.validate_part(type):
                return False

            # If this is a single-part UR then we're done
            if len(components) == 1:
                body = components[0]
                self.result = self.decode_by_type(type, body)
                return True

            # Multi-part URs must have two path components: seq/fragment
            if len(components) != 2:
                raise InvalidPathLength()
            seq = components[0]
            fragment = components[1]

            # Parse the sequence component and the fragment, and make sure they agree.
            (seq_num, seq_len) = URDecoder.parse_sequence_component(seq)
            cbor = Bytewords.decode(Bytewords_Style_minimal, fragment)
            part = FountainEncoderPart.from_cbor(cbor)
            if seq_num != part.seq_num or seq_len != part.seq_len:
                return False

            # Process the part
            if not self.fountain_decoder.receive_part(part):
                return False

            if self.fountain_decoder.is_success():
                self.result = UR(type, self.fountain_decoder.result_message())
            elif self.fountain_decoder.is_failure():
                self.result = self.fountain_decoder.result_error()

            return True
        except Exception:
            return False

    def expected_type(self):
        return self.expected_type

    def expected_part_count(self):
        return self.fountain_decoder.expected_part_count()

    def received_part_indexes(self):
        return self.fountain_decoder.received_part_indexes

    def last_part_indexes(self):
        return self.fountain_decoder.last_part_indexes

    def processed_parts_count(self):
        return self.fountain_decoder.processed_parts_count

    def estimated_percent_complete(self):
        return self.fountain_decoder.estimated_percent_complete()

    def is_success(self):
        result = self.result
        return result if not isinstance(result, Exception) else False

    def is_failure(self):
        result = self.result
        return result if isinstance(result, Exception) else False

    def is_complete(self):
        return self.result != None

    def result_message(self):
        return self.result

    def result_error(self):
        return self.result



#
# bytewords.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .utils import crc32_bytes, partition

BYTEWORDS = "ableacidalsoapexaquaarchatomauntawayaxisbackbaldbarnbeltbetabiasbluebodybragbrewbulbbuzzcalmcashcatschefcityclawcodecolacookcostcruxcurlcuspcyandarkdatadaysdelidicedietdoordowndrawdropdrumdulldutyeacheasyechoedgeepicevenexamexiteyesfactfairfernfigsfilmfishfizzflapflewfluxfoxyfreefrogfuelfundgalagamegeargemsgiftgirlglowgoodgraygrimgurugushgyrohalfhanghardhawkheathelphighhillholyhopehornhutsicedideaidleinchinkyintoirisironitemjadejazzjoinjoltjowljudojugsjumpjunkjurykeepkenokeptkeyskickkilnkingkitekiwiknoblamblavalazyleaflegsliarlimplionlistlogoloudloveluaulucklungmainmanymathmazememomenumeowmildmintmissmonknailnavyneednewsnextnoonnotenumbobeyoboeomitonyxopenovalowlspaidpartpeckplaypluspoempoolposepuffpumapurrquadquizraceramprealredorichroadrockroofrubyruinrunsrustsafesagascarsetssilkskewslotsoapsolosongstubsurfswantacotasktaxitenttiedtimetinytoiltombtoystriptunatwinuglyundouniturgeuservastveryvetovialvibeviewvisavoidvowswallwandwarmwaspwavewaxywebswhatwhenwhizwolfworkyankyawnyellyogayurtzapszerozestzinczonezoom"
WORD_ARRAY = None


def decode_word(word, word_len):
    global WORD_ARRAY
    global BYTEWORDS

    if len(word) != word_len:
        raise ValueError("Invalid Bytewords.")

    dim = 26

    # Since the first and last letters of each Byteword are unique,
    # we can use them as indexes into a two-dimensional lookup table.
    # This table is generated lazily.
    if WORD_ARRAY == None:
        WORD_ARRAY = [-1] * (dim * dim)  # create empty array

        for i in range(256):
            byteword_offset = i * 4
            x = ord(BYTEWORDS[byteword_offset]) - ord("a")
            y = ord(BYTEWORDS[byteword_offset + 3]) - ord("a")
            array_offset = y * dim + x
            WORD_ARRAY[array_offset] = i

    # If the coordinates generated by the first and last letters are out of bounds,
    # or the lookup table contains -1 at the coordinates, then the word is not valid.
    x = ord(word[0].lower()) - ord("a")
    y = ord((word[3 if len(word) == 4 else 1]).lower()) - ord("a")
    if not (0 <= x and x < dim and 0 <= y and y < dim):
        raise ValueError("Invalid Bytewords.")

    offset = y * dim + x
    value = WORD_ARRAY[offset]
    if value == -1:
        raise ValueError("Invalid Bytewords.")

    # If we're decoding a full four-letter word, verify that the two middle letters are correct.
    if len(word) == 4:
        byteword_offset = value * 4
        c1 = word[1].lower()
        c2 = word[2].lower()
        if c1 != BYTEWORDS[byteword_offset + 1] or c2 != BYTEWORDS[byteword_offset + 2]:
            raise ValueError("Invalid Bytewords.")

    # Successful decode.
    return value


def get_word(index):
    byteword_offset = index * 4
    return BYTEWORDS[byteword_offset : byteword_offset + 4]


def get_minimal_word(index):
    byteword_offset = index * 4
    return BYTEWORDS[byteword_offset] + BYTEWORDS[byteword_offset + 3]


def encode(buf, separator):
    words = []
    for i in range(len(buf)):
        byte = buf[i]
        words.append(get_word(byte))

    return separator.join(words)


def add_crc(buf):
    crc_buf = crc32_bytes(buf)
    return buf + crc_buf


def encode_with_separator(buf, separator):
    crc_buf = add_crc(buf)
    return encode(crc_buf, separator)


def encode_minimal(buf):
    result = ""

    crc_buf = add_crc(buf)
    for i in range(len(crc_buf)):
        byte = crc_buf[i]
        result += get_minimal_word(byte)

    return result


def decode(s, separator, word_len):
    buf = bytearray()

    if word_len == 4:
        words = s.split(separator)
    else:
        words = partition(s, 2)

    for word in words:
        buf.append(decode_word(word, word_len))

    if len(buf) < 5:
        raise ValueError("Invalid Bytewords.")

    # Validate checksum
    body = buf[0:-4]
    body_checksum = buf[-4:]
    checksum = crc32_bytes(body)
    if checksum != body_checksum:
        raise ValueError("Invalid Bytewords.")

    return body


Bytewords_Style_standard = 1
Bytewords_Style_uri = 2
Bytewords_Style_minimal = 3


class Bytewords:
    @staticmethod
    def encode(style, bytes):
        if style == Bytewords_Style_standard:
            return encode_with_separator(bytes, " ")
        elif style == Bytewords_Style_uri:
            return encode_with_separator(bytes, "-")
        elif style == Bytewords_Style_minimal:
            return encode_minimal(bytes)
        else:
            assert False

    @staticmethod
    def decode(style, str):
        if style == Bytewords_Style_standard:
            return decode(str, " ", 4)
        elif style == Bytewords_Style_uri:
            return decode(str, "-", 4)
        elif style == Bytewords_Style_minimal:
            return decode(str, 0, 2)
        else:
            assert False



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ...urtypes import RegistryItem
from ...urtypes.cbor import DataItem
from .hd_key import HDKey, CRYPTO_HDKEY
from .ec_key import ECKey, CRYPTO_ECKEY


class MultiKey(RegistryItem):
    def __init__(self, threshold, ec_keys, hd_keys):
        super().__init__()
        self.threshold = threshold
        self.ec_keys = ec_keys
        self.hd_keys = hd_keys

    def __eq__(self, o):
        return self.threshold == o.threshold and self.ec_keys == o.ec_keys and self.hd_keys == o.hd_keys

    @classmethod
    def registry_type(cls):
        return None

    def to_data_item(self):
        map = {}
        map[1] = self.threshold
        combined_keys = self.ec_keys[:] + self.hd_keys[:]
        keys = []
        for key in combined_keys:
            keys.append(DataItem(key.registry_type().tag, key.to_data_item()))
        map[2] = keys
        return map

    @classmethod
    def from_data_item(cls, item):
        map = item.map
        threshold = map[1]
        keys = map[2]
        ec_keys = []
        hd_keys = []
        for key in keys:
            if key.tag == CRYPTO_HDKEY.tag:
                hd_keys.append(HDKey.from_data_item(key))
            elif key.tag == CRYPTO_ECKEY.tag:
                ec_keys.append(ECKey.from_data_item(key))
        return cls(threshold, ec_keys, hd_keys)



#
# crc32.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

# From: https://bitbucket.org/isode/cbor-lite/raw/6c770624a97e3229e3f200be092c1b9c70a60ef1/include/cbor-lite/codec.h

# This file is part of CBOR-lite which is copyright Isode Limited
# and others and released under a MIT license. For details, see the
# COPYRIGHT.md file in the top-level folder of the CBOR-lite software
# distribution.


def bit_length(n):
    return len(bin(abs(n))) - 2


Flag_None = 0
Flag_Require_Minimal_Encoding = 1

Tag_Major_unsignedInteger = 0
Tag_Major_negativeInteger = 1 << 5
Tag_Major_byteString = 2 << 5
Tag_Major_textString = 3 << 5
Tag_Major_array = 4 << 5
Tag_Major_map = 5 << 5
Tag_Major_semantic = 6 << 5
Tag_Major_floatingPoint = 7 << 5
Tag_Major_simple = 7 << 5
Tag_Major_mask = 0xE0

Tag_Minor_length1 = 24
Tag_Minor_length2 = 25
Tag_Minor_length4 = 26
Tag_Minor_length8 = 27

Tag_Minor_false = 20
Tag_Minor_true = 21
Tag_Minor_null = 22
Tag_Minor_undefined = 23
Tag_Minor_half_float = 25
Tag_Minor_singleFloat = 26
Tag_Minor_doubleFloat = 27

Tag_Minor_dateTime = 0
Tag_Minor_epochDateTime = 1
Tag_Minor_positiveBignum = 2
Tag_Minor_negativeBignum = 3
Tag_Minor_decimalFraction = 4
Tag_Minor_bigFloat = 5
Tag_Minor_convertBase64Url = 21
Tag_Minor_convertBase64 = 22
Tag_Minor_convertBase16 = 23
Tag_Minor_cborEncodedData = 24
Tag_Minor_uri = 32
Tag_Minor_base64Url = 33
Tag_Minor_base64 = 34
Tag_Minor_regex = 35
Tag_Minor_mimeMessage = 36
Tag_Minor_selfDescribeCbor = 55799
Tag_Minor_mask = 0x1F
Tag_Undefined = Tag_Major_semantic + Tag_Minor_undefined


def get_byte_length(value):
    if value < 24:
        return 0

    return (bit_length(value) + 7) // 8


class CBOREncoder:
    def __init__(self):
        self.buf = bytearray()

    def get_bytes(self):
        return self.buf

    def encodeTagAndAdditional(self, tag, additional):
        self.buf.append(tag + additional)
        return 1

    def encodeTagAndValue(self, tag, value):
        length = get_byte_length(value)

        # 5-8 bytes required, use 8 bytes
        if length >= 5 and length <= 8:
            self.encodeTagAndAdditional(tag, Tag_Minor_length8)
            self.buf.append((value >> 56) & 0xFF)
            self.buf.append((value >> 48) & 0xFF)
            self.buf.append((value >> 40) & 0xFF)
            self.buf.append((value >> 32) & 0xFF)
            self.buf.append((value >> 24) & 0xFF)
            self.buf.append((value >> 16) & 0xFF)
            self.buf.append((value >> 8) & 0xFF)
            self.buf.append(value & 0xFF)

        # 3-4 bytes required, use 4 bytes
        elif length == 3 or length == 4:
            self.encodeTagAndAdditional(tag, Tag_Minor_length4)
            self.buf.append((value >> 24) & 0xFF)
            self.buf.append((value >> 16) & 0xFF)
            self.buf.append((value >> 8) & 0xFF)
            self.buf.append(value & 0xFF)

        elif length == 2:
            self.encodeTagAndAdditional(tag, Tag_Minor_length2)
            self.buf.append((value >> 8) & 0xFF)
            self.buf.append(value & 0xFF)

        elif length == 1:
            self.encodeTagAndAdditional(tag, Tag_Minor_length1)
            self.buf.append(value & 0xFF)

        elif length == 0:
            self.encodeTagAndAdditional(tag, value)

        else:
            raise Exception("Unsupported byte length of {} for value in encodeTagAndValue()".format(length))

        encoded_size = 1 + length
        return encoded_size

    def encodeUnsigned(self, value):
        return self.encodeTagAndValue(Tag_Major_unsignedInteger, value)

    def encodeNegative(self, value):
        return self.encodeTagAndValue(Tag_Major_negativeInteger, value)

    def encodeInteger(self, value):
        if value >= 0:
            return self.encodeUnsigned(value)
        else:
            return self.encodeNegative(value)

    def encodeBool(self, value):
        return self.encodeTagAndValue(Tag_Major_simple, Tag_Minor_true if value else Tag_Minor_false)

    def encodeBytes(self, value):
        length = self.encodeTagAndValue(Tag_Major_byteString, len(value))
        self.buf += value
        return length + len(value)

    def encodeEncodedBytesPrefix(self, value):
        length = self.encodeTagAndValue(Tag_Major_semantic, Tag_Minor_cborEncodedData)
        return length + self.encodeTagAndAdditional

    def encodeEncodedBytes(self, value):
        length = self.encodeTagAndValue(Tag_Major_semantic, Tag_Minor_cborEncodedData)
        return length + self.encodeBytes(value)

    def encodeText(self, value):
        str_len = len(value)
        length = self.encodeTagAndValue(Tag_Major_textString, str_len)
        self.buf.append(bytes(value, "utf8"))
        return length + str_len

    def encodeArraySize(self, value):
        return self.encodeTagAndValue(Tag_Major_array, value)

    def encodeMapSize(self, value):
        return self.encodeTagAndValue(Tag_Major_map, value)


class CBORDecoder:
    def __init__(self, buf):
        self.buf = buf
        self.pos = 0

    def decodeTagAndAdditional(self, flags=Flag_None):
        if self.pos == len(self.buf):
            raise Exception("Not enough input")
        octet = self.buf[self.pos]
        self.pos += 1
        tag = octet & Tag_Major_mask
        additional = octet & Tag_Minor_mask
        return (tag, additional, 1)

    def decodeTagAndValue(self, flags):
        end = len(self.buf)

        if self.pos == end:
            raise Exception("Not enough input")

        (tag, additional, length) = self.decodeTagAndAdditional(flags)
        if additional < Tag_Minor_length1:
            value = additional
            return (tag, value, length)

        value = 0
        if additional == Tag_Minor_length8:
            if end - self.pos < 8:
                raise Exception("Not enough input")
            for shift in [56, 48, 40, 32, 24, 16, 8, 0]:
                value |= self.buf[self.pos] << shift
                self.pos += 1
            if (flags & Flag_Require_Minimal_Encoding) and value == 0:
                raise Exception("Encoding not minimal")
            return (tag, value, self.pos)
        elif additional == Tag_Minor_length4:
            if end - self.pos < 4:
                raise Exception("Not enough input")
            for shift in [24, 16, 8, 0]:
                value |= self.buf[self.pos] << shift
                self.pos += 1
            if (flags & Flag_Require_Minimal_Encoding) and value == 0:
                raise Exception("Encoding not minimal")
            return (tag, value, self.pos)
        elif additional == Tag_Minor_length2:
            if end - self.pos < 2:
                raise Exception("Not enough input")
            for shift in [8, 0]:
                value |= self.buf[self.pos] << shift
                self.pos += 1
            if (flags & Flag_Require_Minimal_Encoding) and value == 0:
                raise Exception("Encoding not minimal")
            return (tag, value, self.pos)
        elif additional == Tag_Minor_length1:
            if end - self.pos < 1:
                raise Exception("Not enough input")
            value |= self.buf[self.pos]
            self.pos += 1
            if (flags & Flag_Require_Minimal_Encoding) and value == 0:
                raise Exception("Encoding not minimal")
            return (tag, value, self.pos)

        raise Exception("Bad additional value")

    def decodeUnsigned(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_unsignedInteger:
            raise Exception(
                "Expected Tag_Major_unsignedInteger ({}), but found {}".format(Tag_Major_unsignedInteger, tag)
            )
        return (value, length)

    def decodeNegative(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_negativeInteger:
            raise Exception("Expected Tag_Major_negativeInteger, but found {}".format(tag))
        return (value, length)

    def decodeInteger(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)
        if tag == Tag_Major_unsignedInteger:
            return (value, length)
        elif tag == Tag_Major_negativeInteger:
            return (
                -1 - value,
                length,
            )  # TODO: Check that this is the right way -- do we need to use struct.unpack()?

    def decodeBool(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)
        if tag == Tag_Major_simple:
            if value == Tag_Minor_true:
                return (True, length)
            elif value == Tag_Minor_false:
                return (False, length)
            raise Exception("Not a Boolean")
        raise Exception("Not Simple/Boolean")

    def decodeBytes(self, flags=Flag_None):
        # First value is the length of the bytes that follow
        (tag, byte_length, size_length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_byteString:
            raise Exception("Not a byteString")

        end = len(self.buf)
        if end - self.pos < byte_length:
            raise Exception("Not enough input")

        value = bytes(self.buf[self.pos : self.pos + byte_length])
        self.pos += byte_length
        return (value, size_length + byte_length)

    def decodeEncodedBytesPrefix(self, flags=Flag_None):
        (tag, value, length1) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_semantic or value != Tag_Minor_cborEncodedData:
            raise Exception("Not CBOR Encoded Data")

        (tag, value, length2) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_byteString:
            raise Exception("Not byteString")

        return (tag, value, length1 + length2)

    def decodeEncodedBytes(self, flags=Flag_None):
        (tag, minor_tag, tag_length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_semantic or minor_tag != Tag_Minor_cborEncodedData:
            raise Exception("Not CBOR Encoded Data")

        (value, length) = self.decodeBytes(flags)
        return (value, tag_length + length)

    def decodeText(self, flags=Flag_None):
        # First value is the length of the bytes that follow
        (tag, byte_length, size_length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_textString:
            raise Exception("Not a textString")

        end = len(self.buf)
        if end - self.pos < byte_length:
            raise Exception("Not enough input")

        value = bytes(self.buf[self.pos : self.pos + byte_length])
        self.pos += byte_length
        return (value, size_length + byte_length)

    def decodeArraySize(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)

        if tag != Tag_Major_array:
            raise Exception("Expected Tag_Major_array, but found {}".format(tag))
        return (value, length)

    def decodeMapSize(self, flags=Flag_None):
        (tag, value, length) = self.decodeTagAndValue(flags)
        if tag != Tag_Major_mask:
            raise Exception("Expected Tag_Major_map, but found {}".format(tag))
        return (value, length)



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ...urtypes import RegistryType, RegistryItem

CRYPTO_KEYPATH = RegistryType("crypto-keypath", 304)


class Keypath(RegistryItem):
    def __init__(self, components, source_fingerprint, depth):
        super().__init__()
        self.components = components
        self.source_fingerprint = source_fingerprint
        self.depth = depth

    def __eq__(self, o):
        return (
            self.components == o.components
            and self.source_fingerprint == o.source_fingerprint
            and self.depth == o.depth
        )

    @classmethod
    def registry_type(cls):
        return CRYPTO_KEYPATH

    def path(self):
        if not self.components:
            return ""
        return "/".join(
            [
                ("*" if component.wildcard else str(component.index)) + ("'" if component.hardened else "")
                for component in self.components
            ]
        )

    def to_data_item(self):
        map = {}
        components = []
        for component in self.components:
            if component.wildcard:
                components.append([])
            else:
                components.append(component.index)
            components.append(component.hardened)
        map[1] = components
        if self.source_fingerprint is not None:
            map[2] = int.from_bytes(self.source_fingerprint, "big")
        if self.depth is not None:
            map[3] = self.depth
        return map

    @classmethod
    def from_data_item(cls, item):
        map = cls.mapping(item)
        path_components = []
        components = map[1]
        if components:
            for i in range(0, len(components), 2):
                hardened = components[i + 1]
                path = components[i]
                if isinstance(path, int):
                    path_components.append(PathComponent(path, hardened))
                else:
                    path_components.append(PathComponent(None, hardened))
        source_fingerprint = map[2].to_bytes(4, "big") if 2 in map else None
        depth = map[3] if 3 in map else None
        return cls(path_components, source_fingerprint, depth)


class PathComponent:
    def __init__(self, index, hardened):
        self.index = index
        self.hardened = hardened
        self.wildcard = self.index is None
        if self.index and self.index & 0x80000000 != 0:
            raise ValueError("Invalid index - most significant bit cannot be set")

    def __eq__(self, o):
        return self.index == o.index and self.hardened == o.hardened and self.wildcard == o.wildcard



#
# fountain_utils.py
#
# Copyright © 2020 Foundation Devices, Inc.
# Licensed under the "BSD-2-Clause Plus Patent License"
#

from .random_sampler import RandomSampler
from .utils import int_to_bytes
from .xoshiro256 import Xoshiro256

# Fisher-Yates shuffle
def shuffled(items, rng):
    remaining = items
    result = []
    while len(remaining) > 0:
        index = rng.next_int(0, len(remaining) - 1)
        item = remaining.pop(index)
        result.append(item)

    return result


def choose_degree(seq_len, rng):
    degree_probabilities = []
    for i in range(1, seq_len + 1):
        degree_probabilities.append(1.0 / i)

    degree_chooser = RandomSampler(degree_probabilities)
    return degree_chooser.next(lambda: rng.next_double()) + 1


def choose_fragments(seq_num, seq_len, checksum):
    # The first `seq_len` parts are the "pure" fragments, not mixed with any
    # others. This means that if you only generate the first `seq_len` parts,
    # then you have all the parts you need to decode the message.
    if seq_num <= seq_len:
        return set([seq_num - 1])
    else:
        seed = int_to_bytes(seq_num) + int_to_bytes(checksum)
        rng = Xoshiro256.from_bytes(seed)
        degree = choose_degree(seq_len, rng)
        indexes = []

        for i in range(seq_len):
            indexes.append(i)
        shuffled_indexes = shuffled(indexes, rng)
        return set(shuffled_indexes[0:degree])


def contains(set_or_list, el):
    return el in set_or_list


def is_strict_subset(a, b):
    return a.issubset(b)


def set_difference(a, b):
    return a.difference(b)



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io
from ...urtypes import RegistryType, RegistryItem
from ...urtypes.cbor import DataItem
from .multi_key import MultiKey
from .hd_key import HDKey, CRYPTO_HDKEY
from .ec_key import ECKey


class ScriptExpression:
    def __init__(self, tag, expression):
        self.tag = tag
        self.expression = expression

    def __eq__(self, o):
        return self.tag == o.tag and self.expression == o.expression


SCRIPT_EXPRESSION_TAG_MAP = {
    307: ScriptExpression(307, "addr"),
    400: ScriptExpression(400, "sh"),
    401: ScriptExpression(401, "wsh"),
    402: ScriptExpression(402, "pk"),
    403: ScriptExpression(403, "pkh"),
    404: ScriptExpression(404, "wpkh"),
    405: ScriptExpression(405, "combo"),
    406: ScriptExpression(406, "multi"),
    407: ScriptExpression(407, "sortedmulti"),
    408: ScriptExpression(408, "raw"),
    409: ScriptExpression(409, "tr"),
    410: ScriptExpression(410, "cosigner"),
}

CRYPTO_OUTPUT = RegistryType("crypto-output", 308)


class Output(RegistryItem):
    def __init__(self, script_expressions, crypto_key):
        super().__init__()
        self.script_expressions = script_expressions
        self.crypto_key = crypto_key

    def __eq__(self, o):
        return self.script_expressions == o.script_expressions and self.crypto_key == o.crypto_key

    @classmethod
    def registry_type(cls):
        return CRYPTO_OUTPUT

    def descriptor(self, include_checksum=True):
        descriptor = io.StringIO()

        for script_expression in self.script_expressions:
            descriptor.write(script_expression.expression + "(")

        if isinstance(self.crypto_key, MultiKey):
            descriptor.write(str(self.crypto_key.threshold) + ",")

        keys = (
            self.crypto_key.ec_keys[:] + self.crypto_key.hd_keys[:]
            if isinstance(self.crypto_key, MultiKey)
            else [self.crypto_key]
        )
        descriptor.write(",".join([key.descriptor_key() for key in keys]))

        for _ in self.script_expressions:
            descriptor.write(")")

        d = descriptor.getvalue()
        descriptor.close()

        if include_checksum:
            return d + "#" + descriptor_checksum(d)
        return d

    def hd_key(self):
        if isinstance(self.crypto_key, HDKey):
            return self.crypto_key
        return None

    def ec_key(self):
        if isinstance(self.crypto_key, ECKey):
            return self.crypto_key
        return None

    def multi_key(self):
        if isinstance(self.crypto_key, MultiKey):
            return self.crypto_key
        return None

    def to_data_item(self):
        item = DataItem(None, self.crypto_key.to_data_item())
        if self.crypto_key.registry_type() is not None:
            item.tag = self.crypto_key.registry_type().tag
        i = len(self.script_expressions) - 1
        while i >= 0:
            expression = self.script_expressions[i]
            if item.tag is None:
                item.tag = expression.tag
            else:
                item = DataItem(expression.tag, item)
            i -= 1
        return item

    @classmethod
    def from_data_item(cls, item):
        tmp_item = cls.mapping(item)
        script_expressions = []
        while True:
            tag = tmp_item.tag
            if tag in SCRIPT_EXPRESSION_TAG_MAP:
                script_expressions.append(SCRIPT_EXPRESSION_TAG_MAP[tag])
                if isinstance(tmp_item.map, DataItem):
                    tmp_item = tmp_item.map
                else:
                    break
            else:
                break
        exp_len = len(script_expressions)
        is_multi_key = exp_len > 0 and (
            script_expressions[exp_len - 1].expression == "multi"
            or script_expressions[exp_len - 1].expression == "sortedmulti"
        )
        if is_multi_key:
            return cls(script_expressions, MultiKey.from_data_item(tmp_item))

        if tmp_item.tag == CRYPTO_HDKEY.tag:
            return cls(script_expressions, HDKey.from_data_item(tmp_item))
        else:
            return cls(script_expressions, ECKey.from_data_item(tmp_item))


def polymod(c, val):
    c0 = c >> 35
    c = ((c & 0x7FFFFFFFF) << 5) ^ val
    if c0 & 1:
        c ^= 0xF5DEE51989
    if c0 & 2:
        c ^= 0xA9FDCA3312
    if c0 & 4:
        c ^= 0x1BAB10E32D
    if c0 & 8:
        c ^= 0x3706B1677A
    if c0 & 16:
        c ^= 0x644D626FFD
    return c


INPUT_CHARSET = (
    "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
)
CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def descriptor_checksum(descriptor):
    c = 1
    cls = 0
    clscount = 0
    for ch in descriptor:
        pos = INPUT_CHARSET.find(ch)
        if pos == -1:
            return ""
        c = polymod(c, pos & 31)
        cls = cls * 3 + (pos >> 5)
        clscount += 1
        if clscount == 3:
            c = polymod(c, cls)
            cls = 0
            clscount = 0
    if clscount > 0:
        c = polymod(c, cls)
    for _ in range(8):
        c = polymod(c, 0)
    c ^= 1
    checksum = ""
    for i in range(8):
        checksum += CHECKSUM_CHARSET[(c >> (5 * (7 - i))) & 31]
    return checksum



import logging
import bdkpython as bdk

logger = logging.getLogger(__name__)


# Character sets used in the functions
INPUT_CHARSET = (
    "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
)
CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"

# Generators for the polymod function
GENERATOR = [0xF5DEE51989, 0xA9FDCA3312, 0x1BAB10E32D, 0x3706B1677A, 0x644D626FFD]


def descsum_polymod(symbols):
    """Internal function that computes the descriptor checksum."""
    chk = 1
    for value in symbols:
        top = chk >> 35
        chk = (chk & 0x7FFFFFFFF) << 5 ^ value
        for i in range(5):
            chk ^= GENERATOR[i] if ((top >> i) & 1) else 0
    return chk


def descsum_expand(s):
    """Internal function that does the character to symbol expansion"""
    groups = []
    symbols = []
    for c in s:
        if not c in INPUT_CHARSET:
            return None
        v = INPUT_CHARSET.find(c)
        symbols.append(v & 31)
        groups.append(v >> 5)
        if len(groups) == 3:
            symbols.append(groups[0] * 9 + groups[1] * 3 + groups[2])
            groups = []
    if len(groups) == 1:
        symbols.append(groups[0])
    elif len(groups) == 2:
        symbols.append(groups[0] * 3 + groups[1])
    return symbols


def descsum_check(s):
    """Verify that the checksum is correct in a descriptor"""
    if s[-9] != "#" or not all(x in CHECKSUM_CHARSET for x in s[-8:]):
        return False
    symbols = descsum_expand(s[:-9]) + [CHECKSUM_CHARSET.find(x) for x in s[-8:]]
    return descsum_polymod(symbols) == 1


def get_checksum_of_descriptor(s):
    """Add a checksum to a descriptor without"""
    symbols = descsum_expand(s) + [0, 0, 0, 0, 0, 0, 0, 0]
    checksum = descsum_polymod(symbols) ^ 1
    return "".join(CHECKSUM_CHARSET[(checksum >> (5 * (7 - i))) & 31] for i in range(8))


def is_valid_descriptor_checksum(full_descriptor):
    """Check if the checksum of a Bitcoin output descriptor is correct."""
    parts = full_descriptor.split("#")
    if len(parts) != 2:
        raise ValueError("Invalid descriptor format")
    descriptor, provided_checksum = parts
    computed_checksum = get_checksum_of_descriptor(descriptor)
    return computed_checksum == provided_checksum


def has_checksum(descriptor):
    """Check if the Bitcoin output descriptor has a checksum."""
    parts = descriptor.split("#")
    return len(parts) == 2 and len(parts[1]) == 8


def add_checksum_to_descriptor(descriptor_str):
    """Add the correct checksum to a Bitcoin output descriptor."""
    if has_checksum(descriptor_str):
        if is_valid_descriptor_checksum(descriptor_str):
            return descriptor_str
        else:
            raise ValueError("Invalid Checksum already in descriptor")

    checksum = get_checksum_of_descriptor(descriptor_str)
    return f"{descriptor_str}#{checksum}"


def strip_checksum(descriptor_str: str):
    parts = descriptor_str.split("#")
    assert len(parts) == 2 and len(parts[1]) == 8
    return parts[0]


def replace_in_descriptor(descriptor_str, search_str, replace_str):
    assert is_valid_descriptor_checksum(descriptor_str), "Checksum is not valid"
    return add_checksum_to_descriptor(strip_checksum(descriptor_str).replace(search_str, replace_str))


def split_wallet_descriptor(multipath_descriptor_str: str):
    logger.warning(
        "This function is unsafe and must be replaced by bdk/rust miniscript. See https://github.com/bitcoindevkit/bdk/issues/1021"
    )
    assert is_valid_descriptor_checksum(multipath_descriptor_str)

    assert "/<0;1>/*" in multipath_descriptor_str

    return [
        replace_in_descriptor(multipath_descriptor_str, "/<0;1>/*", replace_str)
        for replace_str in ["/0/*", "/1/*"]
    ]


class MultipathDescriptor:
    """
    Will create main+change BDK single and multisig descriptors, no matter if '/<0;1>/*' or '/0/*' or '/1/*' is specified

    This is a temporary class, that can be removed once https://github.com/bitcoindevkit/bdk/issues/1021  is done.
    """

    def __init__(self, bdk_descriptor: bdk.Descriptor, change_descriptor: bdk.Descriptor) -> None:
        self.bdk_descriptors = [bdk_descriptor, change_descriptor]

    @classmethod
    def from_descriptor_str(cls, descriptor_str: str, network: bdk.Network) -> "MultipathDescriptor":
        def count_closing_brackets(s):
            count = 0
            for char in reversed(s):
                if char == ")":
                    count += 1
                else:
                    break
            return count

        descriptor_str = add_checksum_to_descriptor(descriptor_str)

        # check if the descriptor_str is a combined one:
        if "/<0;1>/*" in descriptor_str:
            receive_descriptor_str, change_descriptor_str = split_wallet_descriptor(descriptor_str)
        elif "/0/*" in descriptor_str or "/1/*" in descriptor_str:
            receive_descriptor_str = replace_in_descriptor(descriptor_str, "/1/*", "/0/*")
            change_descriptor_str = replace_in_descriptor(descriptor_str, "/0/*", "/1/*")
        else:
            # sparrow qr code misses the change derivation path completely

            # check if checksum
            stripped = strip_checksum(descriptor_str)

            count_brackets = count_closing_brackets(stripped)
            receive_descriptor_str = stripped[:-count_brackets] + "/0/*" + stripped[-count_brackets:]
            change_descriptor_str = stripped[:-count_brackets] + "/1/*" + stripped[-count_brackets:]

        return MultipathDescriptor(
            bdk.Descriptor(receive_descriptor_str, network=network),
            bdk.Descriptor(change_descriptor_str, network=network),
        )

    def as_string(self) -> str:
        return self._as_string(only_public=True)

    def as_string_private(self) -> str:
        return self._as_string(only_public=False)

    def _as_string(self, only_public=False) -> str:
        logger.warning(
            "This function is unsafe and must be replaced by bdk/rust miniscript. See https://github.com/bitcoindevkit/bdk/issues/1021"
        )
        assert len(self.bdk_descriptors) == 2

        descriptors = [d.as_string() if only_public else d.as_string_private() for d in self.bdk_descriptors]
        # check that these are really just the paths "/0/*" and "/1/*"
        assert all([strip_checksum(d).count(f"/{i}/*)") == 1 for i, d in enumerate(descriptors)])

        # only take the 1 descriptor, and put the /<0;1>/ in there
        return replace_in_descriptor(descriptors[0], f"/{0}/*", f"/<0;1>/*")



from .qr_qui import VideoWidget
from .bitcoin_qr import *
from PyQt6 import QtWidgets
import pygame, math


class BitcoinVideoWidget(VideoWidget):
    def __init__(
        self,
        result_callback=None,
        close_on_result=True,
        parent=None,
        network=bdk.Network.REGTEST,
    ):
        super().__init__(qr_data_callback=self.qr_data_callback, parent=parent)
        self.network: bdk.Network = network

        self.combo_network = QtWidgets.QComboBox(self)
        self.combo_network.addItems([n.name for n in bdk.Network])
        self.combo_network.setCurrentText(self.network.name)
        self.layout().addWidget(self.combo_network)

        self.result_callback = result_callback
        self.close_on_result = close_on_result

        self.meta_data_handler = MetaDataHandler(self.network)
        self.switch_network([n.name for n in bdk.Network].index(self.network.name))
        self.combo_network.currentIndexChanged.connect(self.switch_network)

    def switch_network(self, idx):
        networks = [n for n in bdk.Network]
        self.meta_data_handler.set_network(networks[idx])

    def qr_data_callback(self, qr_data):
        try:
            self.meta_data_handler.add(qr_data.decode("utf-8"))
        except DecodingException as e:
            logger.warning(str(e))

        if self.meta_data_handler.is_complete():
            if self.close_on_result:
                self.close()
            if self.result_callback:
                self.result_callback(self.meta_data_handler.get_complete_data())

    def draw_pie_progress_bar(self, surface, rect, percentage, color):
        x, y, w, h = rect
        # Calculate the center and radius based on the barcode rectangle
        center_x, center_y = x + w / 2, y + h / 2
        radius = min(w, h) / 2  # Adjust the divisor as needed

        # Draw a filled arc (which is actually a filled pie slice)
        radius = min(w, h) // 4  # Radius should not exceed the rectangle
        start_angle = -math.pi  # For example
        stop_angle = -math.pi - math.pi * 2 * percentage

        # Draw lots of small lines to create the filled arc
        num_segments = 100  # Increase this for a smoother arc
        angle_step = (stop_angle - start_angle) / num_segments
        points = [(center_x, center_y)]
        for i in range(num_segments + 1):
            angle = start_angle + i * angle_step
            point_x = center_x + radius * math.cos(angle)
            point_y = center_y + radius * math.sin(angle)
            points.append((point_x, point_y))

        # Use pygame.draw.polygon to draw the filled arc
        pygame.draw.polygon(surface, (0, 255, 0), points)

    def _on_draw_surface(self, surface, barcode):
        super()._on_draw_surface(surface, barcode)

        estimated_percent_complete = self.meta_data_handler.estimated_percent_complete()
        print(estimated_percent_complete)
        if 0 == estimated_percent_complete:
            return

        self.draw_pie_progress_bar(surface, barcode.rect, estimated_percent_complete, (0, 255, 0))


class DemoBitcoinVideoWidget(BitcoinVideoWidget):
    def __init__(
        self,
        parent=None,
        close_on_result=False,
    ):
        super().__init__(
            result_callback=self.result_callback,
            parent=parent,
            close_on_result=close_on_result,
        )

        self.label_qr = QtWidgets.QTextEdit()

        self.layout().addWidget(self.label_qr)

    def result_callback(self, qr_data):
        self.label_qr.setText(str(qr_data))



from abc import ABC, abstractmethod
from os import fdopen
import re
import urllib.parse
from typing import List, Optional, Tuple, Dict
from decimal import Decimal
import bdkpython as bdk
import base64, json
from .urtypes.crypto import PSBT as UR_PSBT
from .urtypes.crypto import Output as US_OUTPUT
from .ur.ur_decoder import URDecoder
from .urtypes.bytes import Bytes as UR_BYTES
import hashlib
import base58
import logging
import enum

from .multipath_descriptor import MultipathDescriptor
from .ur.fountain_encoder import CBOREncoder
from .ur.ur import UR
from .ur.ur_encoder import UREncoder

BITCOIN_BIP21_URI_SCHEME = "bitcoin"
logger = logging.getLogger(__name__)


def is_bitcoin_address(s):
    if re.search(r"^bitcoin\:.*", s, re.IGNORECASE):
        return True

    try:
        bdk.Address(s)
        return True
    except:
        return False


class InvalidBitcoinURI(Exception):
    pass


def serialized_to_hex(serialized):
    return bytes(serialized).hex()


def hex_to_serialized(hex_string):
    return bytes.fromhex(hex_string)


def decode_bip21_uri(uri: str) -> dict:
    """Raises InvalidBitcoinURI on malformed URI."""
    TOTAL_COIN_SUPPLY_LIMIT_IN_BTC = 21000000
    COIN = 100000000

    if not isinstance(uri, str):
        raise InvalidBitcoinURI(f"expected string, not {repr(uri)}")

    if ":" not in uri:
        if not is_bitcoin_address(uri):
            raise InvalidBitcoinURI("Not a bitcoin address")
        return {"address": uri}

    u = urllib.parse.urlparse(uri)
    if u.scheme.lower() != BITCOIN_BIP21_URI_SCHEME:
        raise InvalidBitcoinURI("Not a bitcoin URI")
    address = u.path

    # python for android fails to parse query
    if address.find("?") > 0:
        address, query = u.path.split("?")
        pq = urllib.parse.parse_qs(query)
    else:
        pq = urllib.parse.parse_qs(u.query)

    for k, v in pq.items():
        if len(v) != 1:
            raise InvalidBitcoinURI(f"Duplicate Key: {repr(k)}")

    out = {k: v[0] for k, v in pq.items()}
    if address:
        if not is_bitcoin_address(address):
            raise InvalidBitcoinURI(f"Invalid bitcoin address: {address}")
        out["address"] = address
    if "amount" in out:
        am = out["amount"]
        try:
            m = re.match(r"([0-9.]+)X([0-9])", am)
            if m:
                amount = Decimal(m.group(1)) * pow(Decimal(10), int(m.group(2)) - 8)
            else:
                amount = Decimal(am) * COIN
            if amount > TOTAL_COIN_SUPPLY_LIMIT_IN_BTC * COIN:
                raise InvalidBitcoinURI(f"amount is out-of-bounds: {amount!r} BTC")
            out["amount"] = int(amount)
        except Exception as e:
            raise InvalidBitcoinURI(f"failed to parse 'amount' field: {repr(e)}") from e
    if "message" in out:
        out["message"] = out["message"]
        out["memo"] = out["message"]
    if "time" in out:
        try:
            out["time"] = int(out["time"])
        except Exception as e:
            raise InvalidBitcoinURI(f"failed to parse 'time' field: {repr(e)}") from e
    if "exp" in out:
        try:
            out["exp"] = int(out["exp"])
        except Exception as e:
            raise InvalidBitcoinURI(f"failed to parse 'exp' field: {repr(e)}") from e
    if "sig" in out:
        try:
            out["sig"] = serialized_to_hex(out["sig"])
        except Exception as e:
            raise InvalidBitcoinURI(f"failed to parse 'sig' field: {repr(e)}") from e

    return out


def is_xpub(s):
    if not s.isalnum():
        return False
    first_four_letters = s[:4]
    return first_four_letters.endswith("pub")


class SignerInfo:
    def __init__(
        self,
        fingerprint: str,
        key_origin: str,
        xpub: str,
        derivation_path: Optional[str] = None,
        name: Optional[str] = None,
        first_address: Optional[str] = None,
    ) -> None:
        self.fingerprint = fingerprint
        self.key_origin = self.format_key_origin(key_origin)
        self.xpub = xpub
        self.derivation_path = derivation_path
        self.name = name
        self.first_address = first_address

    def format_key_origin(self, value):
        assert value.startswith("m/"), "The value must start with m/"
        return value.replace("'", "h")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def __str__(self) -> str:
        return f"{self.__dict__}"

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__


def extract_signer_info(s: str) -> SignerInfo:
    """
    Splits 1 keystore,e.g. "[a42c6dd3/84'/1'/0']xpub/0/*"
    into fingerprint, key_origin, xpub, wallet_path

    It also replaces the "'" into "h"

    It overwrites fingerprint, key_origin, xpub  in default_keystore.
    """

    def key_origin_contains_valid_characters(s):
        # Matches strings that consist of 'h', '/', digits, and optionally ends with a single quote
        return re.fullmatch("[mh/0-9']*", s) is not None

    def extract_groups(string, pattern):
        match = re.match(pattern, string)
        if match is None:
            raise Exception(f"'{string}' does not match the required pattern!")
        return match.groups()

    groups = extract_groups(s, r"\[(.*?)\/(.*?)\](.*?)(\/.*?)?$")

    key_origin = "m/" + groups[1].replace("'", "h")
    # guard against false positive detections
    assert key_origin_contains_valid_characters(key_origin)

    return SignerInfo(
        fingerprint=groups[0],
        key_origin=key_origin,
        xpub=groups[2],
        derivation_path=groups[3],
    )


def is_valid_bitcoin_hash(hash):
    import re

    if re.match("^[a-f0-9]{64}$", hash):
        return True
    else:
        return False


def is_valid_wallet_fingerprint(fingerprint):
    import re

    if re.match("^[a-fA-F0-9]{8}$", fingerprint):
        return True
    else:
        return False


###############  here is the simplified electrum base43 code
def base43_decode(v: str):
    __b43chars = b"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ$*+-./:"
    assert len(__b43chars) == 43
    __b43chars_inv = {v: k for k, v in enumerate(__b43chars)}

    base = 43
    num = 0

    # Remove leading zeros and adjust length
    v = v.lstrip("0")

    # Convert each character to a number using the inverse character map
    for char in v:
        num = num * base + __b43chars_inv[ord(char)]

    # Convert the number to bytes
    return num.to_bytes((num.bit_length() + 7) // 8, "big")


################ here is the slip132 part
### see https://github.com/satoshilabs/slips/blob/master/slip-0132.md


def get_version_bytes(slip132_key):
    """Get the version bytes from a SLIP-132 key."""
    raw_extended_key = base58.b58decode(slip132_key)
    return raw_extended_key[:4]


# Mapping of SLIP-132 version bytes to BIP32 version bytes
version_bytes_map = {
    bytes.fromhex("04b24746"): bytes.fromhex("0488b21e"),  # zpub to xpub
    bytes.fromhex("04b2430c"): bytes.fromhex("0488ade4"),  # zprv to xprv
    bytes.fromhex("049d7cb2"): bytes.fromhex("0488b21e"),  # ypub to xpub
    bytes.fromhex("049d7878"): bytes.fromhex("0488ade4"),  # yprv to xprv
    bytes.fromhex("0295b43f"): bytes.fromhex("0488b21e"),  # Ypub to xpub
    bytes.fromhex("0295b005"): bytes.fromhex("0488ade4"),  # Yprv to xprv
    bytes.fromhex("02aa7ed3"): bytes.fromhex("0488b21e"),  # Zpub to xpub
    bytes.fromhex("02aa7a99"): bytes.fromhex("0488ade4"),  # Zprv to xprv
    bytes.fromhex("045f1cf6"): bytes.fromhex("043587cf"),  # vpub to tpub
    bytes.fromhex("045f18bc"): bytes.fromhex("04358394"),  # vprv to tprv
    bytes.fromhex("044a5262"): bytes.fromhex("043587cf"),  # upub to tpub
    bytes.fromhex("044a4e28"): bytes.fromhex("04358394"),  # uprv to tprv
    bytes.fromhex("024289ef"): bytes.fromhex("043587cf"),  # Upub to tpub
    bytes.fromhex("024285b5"): bytes.fromhex("04358394"),  # Uprv to tprv
    bytes.fromhex("02575483"): bytes.fromhex("043587cf"),  # Vpub to tpub
    bytes.fromhex("02575048"): bytes.fromhex("04358394"),  # Vprv to tprv
}


def base58check_decode(s):
    """Decode a Base58Check encoded string to bytes"""
    # Decode the string
    data = base58.b58decode(s)

    # Split the data into the payload and the checksum
    check_sum = data[-4:]
    payload = data[:-4]

    # Calculate the checksum of the payload
    calculated_check_sum = hashlib.sha256(hashlib.sha256(payload).digest()).digest()[:4]

    # Compare the calculated checksum with the checksum from the data
    if check_sum != calculated_check_sum:
        raise ValueError("Invalid checksum")

    return payload


def convert_slip132_to_bip32(slip132_key):
    """Convert a SLIP-132 extended key to a BIP32 extended key."""
    raw_extended_key = base58check_decode(slip132_key)
    slip132_version_bytes = raw_extended_key[:4]

    # Lookup the corresponding BIP32 version bytes
    bip32_version_bytes = version_bytes_map.get(slip132_version_bytes)
    if bip32_version_bytes is None:
        raise ValueError("Unsupported SLIP-132 version bytes")

    # Replace the version bytes of the raw key
    replaced_version_key = bip32_version_bytes + raw_extended_key[4:]

    # Calculate the checksum of the replaced version key
    check_sum = hashlib.sha256(hashlib.sha256(replaced_version_key).digest()).digest()[:4]

    # Encode the replaced version key + checksum into Base58
    bip32_key = base58.b58encode(replaced_version_key + check_sum).decode()

    return bip32_key


def is_slip132(key):
    return get_version_bytes(key) in version_bytes_map


class DataType(enum.Enum):
    Bip21 = enum.auto()  # https://bips.dev/21/
    Descriptor = enum.auto()
    MultiPathDescriptor = enum.auto()
    Xpub = enum.auto()
    Fingerprint = enum.auto()
    SignerInfo = enum.auto()  # FingerPrint, Derivation path, Xpub
    PSBT = enum.auto()
    Txid = enum.auto()
    Tx = enum.auto()
    SignerInfos = enum.auto()  # a list of SignerInfo


    @classmethod
    def from_value(cls, value: int)-> "DataType":
        min_value = min([member.value for member in cls])
        return list(cls)[value - min_value]

    @classmethod
    def from_name(cls, value: str) -> "DataType":
        names = [member.name for member in cls]
        return list(cls)[names.index(value)]




class DecodingException(Exception):
    pass


class InconsistentDescriptors(Exception):
    pass


class WrongNetwork(Exception):
    pass


class Data:
    """
    Recognized bitcoin data in a string, gives the data and the DataType
    """

    def __init__(self, data, data_type: DataType) -> None:
        self.data = data
        self.data_type = data_type

    def dump(self)->Dict:
        return {"data":self.data_as_string(), "data_type":self.data_type.name}
    
    @classmethod
    def from_dump(cls, d:Dict, network:bdk.Network)->"Data":
        data  = cls.from_str(d["data"],network=network)
        d["data_type"] = DataType.from_name(d["data_type"])
        assert d["data_type"] == data.data_type
        return data

    def data_as_string(self):
        if isinstance(self.data, str):
            return self.data
        if self.data_type == DataType.Bip21:
            return str(self.data)
        if self.data_type == DataType.Descriptor:
            return self.data.as_string_private() if self.data else self.data
        if self.data_type == DataType.MultiPathDescriptor:
            return self.data.as_string_private() if self.data else self.data
        if self.data_type == DataType.SignerInfo:
            return str(self.data)
        if self.data_type == DataType.SignerInfos:
            return str(self.data)
        if self.data_type == DataType.PSBT:
            return str(self.data.serialize())
        if self.data_type == DataType.Tx:
            return str(serialized_to_hex(self.data.serialize()))

        return str(self.data)

    def __str__(self) -> str:
        return f"{self.data_type.name}: {self.data_as_string()}"

    @classmethod
    def _try_decode_bip21(cls, s):
        try:
            return decode_bip21_uri(s)
        except Exception:
            pass

    @classmethod
    def _try_get_descriptor(cls, s, network):
        try:
            descriptor = bdk.Descriptor(s, network)
            if descriptor:
                logger.debug("detected descriptor")
                return descriptor
        except Exception:
            pass

        try:
            specter_dict = json.loads(s)
            if "descriptor" in specter_dict:
                descriptor = bdk.Descriptor(specter_dict["descriptor"], network)
                logger.debug("detected descriptor")
                return descriptor
        except Exception:
            pass

    @classmethod
    def _try_get_multipath_descriptor(cls, s, network):
        # if new lines are presnt, try checking if there are descriptors in the lines
        if "\n" in s:
            splitted_lines = s.split("\n")
            results = [cls._try_get_multipath_descriptor(line.strip(), network) for line in splitted_lines]
            # check that all entries return the same multipath descriptor
            # "None" entries are disallowed, to ensure that no depection is possible
            if all(results):
                all_identical = all(
                    element.as_string_private() == results[0].as_string_private() for element in results
                )
                if all_identical:
                    return results[0]
                else:
                    # the string contins multiple inconsitent descriptors.
                    # Don't know how to handle this properly.
                    raise InconsistentDescriptors(f"The descriptors {splitted_lines} are inconsistent.")
            else:
                # if splitting lines didnt work, then remove the \n and try to recognize all as 1 descriptor
                s = s.replace("\n", "")

        try:
            multipath_descriptor = MultipathDescriptor.from_descriptor_str(s, network)
            if multipath_descriptor:
                logger.debug("detected descriptor")
                return multipath_descriptor
        except Exception:
            pass

    @classmethod
    def _decoding_strategies(cls):
        return [
            lambda x: base64.b64decode(x),  # base64 decoding
            lambda x: bytes.fromhex(x),  # hex decoding
            lambda x: base43_decode(x),  # base43 decoding
            lambda x: base58.b58decode(x),
        ]

    @classmethod
    def _try_decode_psbt(cls, s):
        psbt_magic_bytes = b"psbt\xff"

        # Try each decoding strategy in the loop
        for decode in cls._decoding_strategies():
            try:
                decoded = decode(s)
                if decoded[:5] == psbt_magic_bytes:
                    return bdk.PartiallySignedTransaction(base64.b64encode(decoded).decode())
            except Exception:
                continue  # If one strategy fails, try the next

        return None

    @classmethod
    def _try_decode_serialized_transaction(cls, s):
        # Try each decoding strategy in the loop
        for decode in cls._decoding_strategies():
            try:
                decoded = decode(s)
                return bdk.Transaction(decoded)
            except Exception:
                continue  # If one strategy fails, try the next

        return None

    @classmethod
    def _try_extract_signer_info(cls, s, network):
        signer_info = None
        try:
            signer_info = extract_signer_info(s)
        except:
            pass

        if signer_info:
            logger.debug("detected signer_info")
            if is_slip132(signer_info.xpub):
                signer_info.xpub = convert_slip132_to_bip32(signer_info.xpub)
            return signer_info

        # try to load from a generic json (and cobo)
        try:
            d = json.loads(s)
            fingerprint = None
            key_origin = None
            xpub = None
            if d.get("fingerprint"):
                fingerprint = d.get("fingerprint")
            if d.get("xfp"):
                fingerprint = d.get("xfp")
            if d.get("key_origin"):
                key_origin = d.get("key_origin")
            if d.get("deriv"):
                key_origin = d.get("deriv")
            if d.get("path"):
                # cobo convention
                key_origin = d.get("path")
            if d.get("xpub"):
                xpub = d.get("xpub")
            if fingerprint and key_origin and xpub:
                return SignerInfo(fingerprint=fingerprint, key_origin=key_origin, xpub=xpub)
        except Exception:
            pass

        return None

    @classmethod
    def _try_extract_signer_infos(cls, s, network):
        # if it is a json
        json_data = None
        try:
            json_data = json.loads(s)

            # check if it is in coldcard export format

            assert "chain" in json_data
            assert "xfp" in json_data
            assert "xpub" in json_data

        except:
            return None

        if network == bdk.Network.BITCOIN:
            if json_data["chain"] != "BTC":
                raise WrongNetwork()
        if network == bdk.Network.REGTEST:
            if json_data["chain"] != "XRT":
                raise WrongNetwork()
        if network == bdk.Network.TESTNET:
            if json_data["chain"] != "XTN":
                raise WrongNetwork()
        if network == bdk.Network.SIGNET:
            # unclear which chain value is used for signet in coldcard
            # https://coldcard.com/docs/upgrade/#mk4-version-511-feb-27-2023
            if json_data["chain"] not in ["XTN", "XRT"]:
                raise WrongNetwork()

        # the fingerprint in the top level is the relevant one
        fingerprint = json_data["xfp"]

        return [
            SignerInfo(
                xpub=v["xpub"],
                fingerprint=fingerprint,
                key_origin=v["deriv"],
                first_address=v["first"] if "first" in v else None,
                name=v["name"],
            )
            for k, v in json_data.items()
            if k.lower().startswith("bip")
        ]

    @classmethod
    def from_str(cls, s, network) -> "Data":
        s = s.strip()
        data = None

        # Sequence of checks to identify the type of data in `s`
        if decoded_bip21 := cls._try_decode_bip21(s):
            return Data(decoded_bip21, DataType.Bip21)

        if is_xpub(s):
            data = convert_slip132_to_bip32(s) if is_slip132(s) else s
            return Data(data, DataType.Xpub)

        if descriptor := cls._try_get_descriptor(s, network):
            return Data(descriptor, DataType.Descriptor)

        if descriptor := cls._try_get_multipath_descriptor(s, network):
            return Data(descriptor, DataType.MultiPathDescriptor)

        if is_valid_bitcoin_hash(s):
            return Data(s, DataType.Txid)

        if is_valid_wallet_fingerprint(s):
            return Data(s, DataType.Fingerprint)

        if psbt := cls._try_decode_psbt(s):
            return Data(psbt, DataType.PSBT)

        if tx := cls._try_decode_serialized_transaction(s):
            return Data(tx, DataType.Tx)

        if signer_info := cls._try_extract_signer_info(s, network):
            return Data(signer_info, DataType.SignerInfo)

        if signer_infos := cls._try_extract_signer_infos(s, network):
            return Data(signer_infos, DataType.SignerInfos)

        raise DecodingException(f"{s} Could not be decoded")

    def generate_fragments_for_qr(self, max_qr_size=100):
        serialized = self.data_as_string()

        if len(serialized) <= max_qr_size:
            return [serialized]

        if self.data_type == DataType.Tx:
            bcor_encoder = CBOREncoder()
            bcor_encoder.encodeBytes(hex_to_serialized(serialized))
            ur = UR("bytes", bcor_encoder.get_bytes())
        elif self.data_type == DataType.PSBT:
            bcor_encoder = CBOREncoder()
            bcor_encoder.encodeBytes(base64.b64decode(serialized.encode()))
            ur = UR("crypto-psbt", bcor_encoder.get_bytes())

        encoder = UREncoder(ur, max_fragment_len=max_qr_size)
        fragments = []
        while not encoder.is_complete():
            part = encoder.next_part()
            fragments.append(part)
        return fragments

    def write_to_filedescriptor(self, fd):
        """

        This is useful for a temporary file
        fd, file_path = tempfile.mkstemp(
            suffix=f"_{key}.{self.file_extension}", prefix=""
        )

        a file descriptor for a normal file can be created with
                fd = os.open(filename, os.O_CREAT | os.O_WRONLY)
        """
        if self.data_type == DataType.Tx:
            with fdopen(fd, "wb") as file:
                file.write(bytes(self.data.serialize()))
        elif self.data_type == DataType.PSBT:
            with fdopen(fd, "wb") as file:
                file.write(base64.b64decode(self.data.serialize()))
        else:
            with fdopen(fd, "w") as file:
                file.write(self.data_as_string())


class BaseCollector(ABC):
    def __init__(self, network) -> None:
        self.data: Optional[Data] = None
        self.network = network

    @abstractmethod
    def is_correct_data_format(self, s):
        pass

    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def get_complete_data(self) -> Optional[Data]:
        pass

    @abstractmethod
    def add(self, s: str):
        pass

    def clear(self):
        self.data = None

    @abstractmethod
    def estimated_percent_complete(self):
        pass


class SinglePassCollector(BaseCollector):
    def is_correct_data_format(self, s):
        return True

    def is_complete(self) -> bool:
        return bool(self.data)

    def get_complete_data(self) -> Optional[Data]:
        return self.data

    def add(self, s: str) -> Data:
        self.data = Data.from_str(s, network=self.network)
        return self.data

    def estimated_percent_complete(self):
        return float(bool(self.data))


class SpecterDIYCollector(BaseCollector):
    def __init__(self, network) -> None:
        super().__init__(network)
        self.clear()
        self.total_parts: int = 0

    def is_correct_data_format(self, s):
        return self.extract_specter_diy_qr_part(s) is not None

    def is_complete(self) -> bool:
        return len(self.parts) == self.total_parts

    def get_complete_data(self) -> Optional[Data]:
        if not self.is_complete():
            return None

        total_s = ""
        for i in range(1, self.total_parts + 1):
            total_s += self.parts[i]
        return Data.from_str(total_s, network=self.network)

    def extract_specter_diy_qr_part(self, s) -> Optional[Tuple[int, int, str]]:
        "pMofM something  ->  (M,N,something)"
        pattern = r"^p(\d+)of(\d+)\s(.*)"
        match = re.match(pattern, s)
        if match:
            return int(match.group(1)), int(match.group(2)), match.group(3)
        return None

    def add(self, s: str) -> Optional[str]:
        specter_diy_qr_part = self.extract_specter_diy_qr_part(s)
        if not specter_diy_qr_part:
            return None
        m, n, data = specter_diy_qr_part
        if (self.total_parts is not None) and n != self.total_parts:
            # if n != self.total_parts then, it appears we switched to a different qrcode
            self.clear()

        if self.total_parts is None:
            self.total_parts = n

        self.parts[m] = data
        return data

    def clear(self):
        super().clear()
        self.parts: Dict[int, str] = {}
        self.total_parts = None

    def estimated_percent_complete(self):
        if not self.total_parts:
            return 0
        return min(len(self.parts) / max(self.total_parts, 1), 1)


class URCollector(BaseCollector):
    def __init__(self, network) -> None:
        super().__init__(network)
        self.clear()
        self.last_received_part: Optional[str] = None

    def is_psbt(self, s: str):
        return re.search("^UR:CRYPTO-PSBT/", s, re.IGNORECASE)

    def is_descriptor(self, s: str):
        return re.search("^UR:CRYPTO-OUTPUT/", s, re.IGNORECASE)

    def is_bytes(self, s: str):
        return re.search("^UR:BYTES/", s, re.IGNORECASE)

    def is_correct_data_format(self, s):
        if self.is_psbt(s):
            return True
        if self.is_descriptor(s):
            return True
        if self.is_bytes(s):
            return True

        return False

    def is_complete(self) -> bool:
        return self.decoder.is_complete()

    def get_complete_data(self) -> Data:
        if self.decoder.result.type == "crypto-psbt":
            qr_content = UR_PSBT.from_cbor(self.decoder.result.cbor).data
            s = base64.b64encode(qr_content).decode("utf-8")
        if self.decoder.result.type == "crypto-output":
            s = US_OUTPUT.from_cbor(self.decoder.result.cbor).descriptor()
        if self.decoder.result.type == "bytes":
            raw = UR_BYTES.from_cbor(self.decoder.result.cbor).data
            s = raw.hex()

        return Data.from_str(s, network=self.network)

    def add(self, s: str) -> Optional[str]:
        self.decoder.receive_part(s)
        logger.debug(f"{round(self.decoder.estimated_percent_complete()*100)}% complete")

        # if the decoder is stuck for some reason. reset it
        # this part must be done AFTER receive_part(s)
        if self.received_parts > self.decoder.expected_part_count() * 2:
            logger.debug(
                f"{self.received_parts}/{self.decoder.expected_part_count()} parts received, but incomplete. Resetting"
            )
            self.clear()

        if self.last_received_part != s:
            self.received_parts += 1
            self.last_received_part = s
        return s

    def clear(self):
        super().clear()
        self.decoder = URDecoder()
        self.received_parts = 0
        self.last_received_part = None

    def estimated_percent_complete(self):
        return len(self.decoder.received_part_indexes()) / self.decoder.expected_part_count()


class MetaDataHandler:
    "Unified class to handle animated and static qr codes"

    def __init__(self, network) -> None:
        self.network = network
        # SinglePassCollector must be the last one
        self.collectors: List[BaseCollector] = [
            URCollector(self.network),
            SpecterDIYCollector(self.network),
            SinglePassCollector(self.network),
        ]
        self.last_used_collector: Optional[BaseCollector] = None

    def set_network(self, network: bdk.Network):
        self.network = network
        for collector in self.collectors:
            collector.network = network

    def get_collector(self, s: str) -> Optional[BaseCollector]:
        for collector in self.collectors:
            if collector.is_correct_data_format(s):
                return collector
        return None

    def add(self, s: str):
        self.last_used_collector = self.get_collector(s)
        if not self.last_used_collector:
            raise Exception("Could not get a fitting colletor")
        return self.last_used_collector.add(s)

    def is_complete(self) -> bool:
        if not self.last_used_collector:
            return False
        return self.last_used_collector.is_complete()

    def get_complete_data(self) -> Optional[Data]:
        if not self.last_used_collector:
            return None
        data = self.last_used_collector.get_complete_data()
        self.last_used_collector.clear()
        return data

    def estimated_percent_complete(self):
        if not self.last_used_collector:
            return 0
        return self.last_used_collector.estimated_percent_complete()



# The MIT License (MIT)

# Copyright (c) 2021 Tom J. Sun
# Copyright (c) 2015 Sokolov Yura
# Copyright (c) 2013 Fritz Grimpen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# coding: utf-8

import struct
import math

from .data import DataItem, Undefined


class InvalidCborError(Exception):
    pass


class _Break(InvalidCborError):
    def __init__(self):
        InvalidCborError.__init__(self, "Invalid BREAK code occurred")


class Decoder(object):
    def __init__(self, input):
        self._jump_table = [
            lambda *args: self.decode_integer(*args, sign=False),
            lambda *args: self.decode_integer(*args, sign=True),
            self.decode_bytestring,
            self.decode_textstring,
            self.decode_list,
            self.decode_dict,
            self.decode_tagging,
            self.decode_other,
        ]
        self.input = input

    def decode(self):
        mtype, ainfo = self._decode_ibyte()
        try:
            decoder = self._jump_table[mtype]
        except KeyError:
            raise InvalidCborError("Invalid major type {}".format(mtype))
        return decoder(mtype, ainfo)

    def decode_integer(self, mtype, ainfo, sign=False):
        res = self._decode_length(ainfo)
        if sign is True:
            return -1 - res
        else:
            return res

    def decode_bytestring(self, mtype, ainfo):
        length = self._decode_length(ainfo)
        if length is None:
            res = bytearray()
            while True:
                mtype_, ainfo_ = self._decode_ibyte()
                if (mtype_, ainfo_) == (7, 31):
                    break
                if mtype_ != 2:
                    pass
                res.extend(self.decode_bytestring(mtype_, ainfo_))
            return bytes(res)
        else:
            return self._read(length)

    def decode_textstring(self, mtype, ainfo):
        length = self._decode_length(ainfo)
        if length is None:
            res = bytearray()
            while True:
                mtype_, ainfo_ = self._decode_ibyte()
                if (mtype_, ainfo_) == (7, 31):
                    break
                if mtype_ != 3:
                    pass
                res.extend(self.decode_bytestring(mtype_, ainfo_))
            return res.decode("utf-8")
        else:
            return self._read(length).decode("utf-8")

    def decode_list(self, mtype, ainfo):
        length = self._decode_length(ainfo)
        if length is None:
            res = []
            while True:
                try:
                    res.append(self.decode())
                except _Break:
                    break
            return res
        else:
            res = [None for _ in range(length)]
            for n in range(length):
                res[n] = self.decode()
            return res

    def decode_dict(self, mtype, ainfo):
        length = self._decode_length(ainfo)
        if length is None:
            res = {}
            try:
                while True:
                    key = self.decode()
                    value = self.decode()
                    res[key] = value
            except _Break:
                pass
            return res
        else:
            res = {}
            for n in range(length):
                key, value = self.decode(), self.decode()
                res[key] = value
            return res

    def decode_tagging(self, mtype, ainfo):
        length = self._decode_length(ainfo)
        return DataItem(length, self.decode())

    def decode_half_float(self, mtype, ainfo):
        half = struct.unpack(">H", self._read(2))[0]
        valu = (half & 0x7FFF) << 13 | (half & 0x8000) << 16
        if (half & 0x7C00) != 0x7C00:
            return math.ldexp(struct.unpack("!f", struct.pack("!I", valu))[0], 112)
        return struct.unpack("!f", struct.pack("!I", valu | 0x7F800000))[0]

    def decode_single_float(self, mtype, ainfo):
        return struct.unpack(">f", self._read(4))[0]

    def decode_double_float(self, mtype, ainfo):
        return struct.unpack(">d", self._read(8))[0]

    def decode_other(self, mtype, ainfo):
        if ainfo == 20:
            return False
        elif ainfo == 21:
            return True
        elif ainfo == 22:
            return None
        elif ainfo == 23:
            return Undefined
        elif ainfo == 25:
            return self.decode_half_float(mtype, ainfo)
        elif ainfo == 26:
            return self.decode_single_float(mtype, ainfo)
        elif ainfo == 27:
            return self.decode_double_float(mtype, ainfo)
        elif ainfo == 31:
            raise _Break()

    def _decode_ibyte(self):
        byte = self._read(1)[0]
        if isinstance(byte, str):
            byte = ord(byte)
        return (byte & 0b11100000) >> 5, byte & 0b00011111

    def _decode_length(self, ainfo):
        if ainfo < 24:
            return ainfo
        elif ainfo == 24:
            return from_bytes(self._read(1))
        elif ainfo == 25:
            return from_bytes(self._read(2))
        elif ainfo == 26:
            return from_bytes(self._read(4))
        elif ainfo == 27:
            return from_bytes(self._read(8))
        elif ainfo == 31:
            return None
        raise InvalidCborError("Invalid additional information {}".format(ainfo))

    def _read(self, n):
        m = self.input.read(n)
        if len(m) != n:
            raise InvalidCborError("Expected {} bytes, got {} bytes instead".format(n, len(m)))
        return m


__all__ = ["InvalidCborError", "Decoder"]


def from_bytes(val):
    return int.from_bytes(val, "big")



import pygame
import pygame.camera
from PyQt6 import QtCore, QtWidgets, QtGui
from pyzbar import pyzbar
import time
from PyQt6.QtCore import QEvent
import mss
import numpy as np
from PyQt6.QtCore import pyqtSignal


class VideoWidget(QtWidgets.QWidget):
    signal_qr_data_callback = pyqtSignal(object)

    def __init__(self, qr_data_callback=None, parent=None):
        super().__init__(parent)
        self.is_screen_capture = False
        self.screen_capturer = mss.mss()

        self.label_image = QtWidgets.QLabel()
        self.qr_data_callback = qr_data_callback
        self.signal_qr_data_callback.connect(qr_data_callback)

        pygame.camera.init()
        self.cameras = self.get_valid_cameras()
        self.combo_cameras = QtWidgets.QComboBox()
        self.combo_cameras.addItems(self.cameras)
        self.combo_cameras.currentIndexChanged.connect(self.switch_camera)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label_image)
        layout.addWidget(self.combo_cameras)

        self.setLayout(layout)

        self.capture = None

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_frame)
        self.timer.start(int(1000.0 / 30.0))  # 30 FPS

        # Add "Screen" option for screen capture
        self.combo_cameras.addItem("Screen")
        self.cameras.append("Screen")  # Adding "Screen" to the cameras list
        self.switch_camera(0)

    def closeEvent(self, event: QEvent) -> None:
        self.timer.stop()
        if self.capture:
            try:
                self.capture.stop()
            except:
                pass

        # If you call the parent's closeEvent(), it will proceed to close the widget
        super().closeEvent(event)

    def get_valid_cameras(self):
        valid_cameras = []
        for camera in pygame.camera.list_cameras():
            try:
                temp_camera = pygame.camera.Camera(camera, (640, 480))
                temp_camera.start()
                temp_camera.stop()
                valid_cameras.append(camera)
            except SystemError:
                continue
        return valid_cameras

    def switch_camera(self, index):
        selected_camera = self.combo_cameras.currentText()
        if selected_camera == "Screen":
            self.is_screen_capture = True
            # Additional logic for initializing screen capture can be added here
        else:
            self.is_screen_capture = False

            if self.capture:
                try:
                    self.capture.stop()
                except:
                    pass

            self.capture = pygame.camera.Camera(self.cameras[index], (640, 480))
            self.capture.start()
            time.sleep(0.1)  # Add a short delay

    def _numpy_to_surface(self, numpy_image):
        # Convert numpy image (RGB) to pygame surface
        return pygame.surfarray.make_surface(numpy_image.transpose((1, 0, 2)))

    def _surface_to_numpy(self, surface):
        # Convert pygame surface to numpy array (RGB)
        return pygame.surfarray.array3d(surface).transpose((1, 0, 2))

    def _on_draw_surface(self, surface, barcode):
        x, y, w, h = barcode.rect
        pygame.draw.rect(surface, (0, 255, 0), (x, y, w, h), 2)

    def update_frame(self):

        barcodes = []
        if self.is_screen_capture:
            with mss.mss() as sct:
                monitor = sct.monitors[1]  # capture the first monitor
                sct_img = sct.grab(monitor)
                image = np.array(sct_img)

                # Convert BGRA to RGB
                image = image[:, :, :3]

                # Convert numpy image to surface for drawing
                surface = self._numpy_to_surface(image)
        else:
            try:
                surface = self.capture.get_image()
            except:
                # print("Could not get image")
                return

        barcodes = pyzbar.decode(pygame.surfarray.array3d(surface))
        surface = pygame.transform.flip(surface, False, True)
        surface = pygame.transform.rotate(surface, -90)

        for barcode in barcodes:
            self._on_draw_surface(surface, barcode)
            # only show the 1. barcode
            break

        surface = pygame.transform.flip(surface, False, True)
        self.showSurface(surface, scale_to=(640, 480))

        for barcode in barcodes:
            self.signal_qr_data_callback.emit(barcode.data)
            break

    def showSurface(self, surface, scale_to=(640, 480)):

        array3d = pygame.surfarray.array3d(surface)
        height, width, _ = array3d.shape
        bytes_per_line = 3 * width
        q_image = QtGui.QImage(array3d.data, width, height, bytes_per_line, QtGui.QImage.Format.Format_RGB888)

        # Convert QImage to QPixmap
        pixmap = QtGui.QPixmap.fromImage(q_image)

        # Scale the pixmap if scaling dimensions are provided
        if scale_to:
            pixmap = pixmap.scaled(
                scale_to[0],
                scale_to[1],
                QtCore.Qt.AspectRatioMode.KeepAspectRatio,
                QtCore.Qt.TransformationMode.SmoothTransformation,
            )

        self.label_image.setPixmap(pixmap)


class DemoVideoWidget(VideoWidget):
    def __init__(self, parent=None):
        super().__init__(qr_data_callback=self.show_qr, parent=parent)

        self.label_qr = QtWidgets.QTextEdit()

        self.layout().addWidget(self.label_qr)

    def show_qr(self, qr_data):
        s = qr_data.decode("utf-8")
        print(s)
        self.label_qr.setText(s)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)

    video_widget = DemoVideoWidget()
    video_widget.show()

    sys.exit(app.exec_())
